;;; -*- auto-recompile: t -*-

;;; graphs.el --- basic functions for plotting graphs

;; This file is not part of GNU Emacs.
;; 
;; csound-ft.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; csound-ft.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Author: Stphane Rollandin <hepta@zogotounga.net>
;;---------

(defun make-graph (nx ny)
  (let ((gv (make-vector ny nil)))
    (dotimes (vline ny gv)
      (aset gv vline (make-bool-vector nx t)))))

(defun draw-in-graph (graph x y)
  (condition-case nil
    (aset (aref graph y) x nil)
    (error nil)))

(defun insert-graph (graph &optional foreground background)
  (insert-image `(image :type xbm :data ,graph :width 500 :height 200 :foreground  ,(or foreground "black") :background ,(or background "white"))))

(defun make-plotting-area (nx ny x0 x1 y0 y1)
  `(plotting-area
    :width ,nx
    :height ,ny
    :graph ,(make-graph nx ny)
    :x0 ,x0
    :y0 ,y0
    :x1 ,x1
    :y1 ,y1
    :dx  ,(/ (- x1 x0) (* nx 1.0))
    :dy  ,(/ (- y1 y0) (* ny 1.0))
    :x (lambda (x) (+ ,x0 (* x ,(/ (- x1 x0) (* nx 1.0)))))
    :y (lambda (y) (+ ,y0 (* (- ,ny y) ,(/ (- y1 y0) (* ny 1.0)))))
    :nx (lambda (x) (round (* (- x ,x0) ,(/ (* nx 1.0) (- x1 x0)))))
    :ny (lambda (y) (round (* (- ,y1 y) ,(/ (* ny 1.0) (- y1 y0)))))))

(defun plot-point (x y parea)
  (draw-in-graph 
   (plist-get (cdr parea) :graph)
   (funcall (plist-get (cdr parea) :nx) x)
   (funcall (plist-get (cdr parea) :ny) y)))

(defun plot-segment (x0 y0 x1 y1 parea)
  (let* ((graph (plist-get (cdr parea) :graph))
	 (nx0 (funcall (plist-get (cdr parea) :nx) x0))
	 (ny0 (funcall (plist-get (cdr parea) :ny) y0))
	 (nx1 (funcall (plist-get (cdr parea) :nx) x1))
	 (ny1 (funcall (plist-get (cdr parea) :ny) y1))
	 (mdn (max (abs (- nx1 nx0)) (abs (- ny1 ny0)))))
    (if (<= mdn 1)
	(draw-in-graph graph nx0 ny0)
      (dotimes (n mdn)
	(let ((x (round (+ nx0 (* n (/ (- nx1 nx0) (* mdn 1.0))))))
	      (y (round (+ ny0 (* n (/ (- ny1 ny0) (* mdn 1.0)))))))
	  (draw-in-graph graph x y))))))

(defun insert-plotting-area (parea &optional foreground background mouse-1-action)
  (insert-image `(image :type xbm 
			:data ,(plist-get (cdr parea) :graph) 
			:width ,(plist-get (cdr parea) :width) 
			:height ,(plist-get (cdr parea) :height)
			:foreground  ,(or foreground "black") 
			:background ,(or background "white"))
		(propertize " "
			    'local-map  (let ((map (make-sparse-keymap)))
					  (define-key map [mouse-1] mouse-1-action)
					  map))))

(defun plot-function (f parea &optional segmented)
  (let ((x (plist-get (cdr parea) :x0))
	(dx (plist-get (cdr parea) :dx)))
    (dotimes (nx (plist-get (cdr parea) :width))
      (condition-case nil
	  (if segmented
	      (plot-segment x (funcall f x) (+ x dx) (funcall f (+ x dx)) parea)        ;;  optimiser !!
	    (plot-point x (funcall f x) parea))
	  (error nil))
      (incf x dx))))

(defun plot-vector (v parea &optional segmented)
  (plot-function (lambda (x) (aref v (round x))) parea segmented)
;  (setf (cdr parea) (plist-put (cdr parea) :vector v))
)

(defun pop-up-area (parea &optional foreground background text title mouse-1-action)
  (let* ((pop-up-frames t)
	 (frame (window-frame (display-buffer (set-buffer (generate-new-buffer "*PLOT*")))))
	 (fw (/ (plist-get (cdr parea) :width) (frame-char-width frame)))
	 (fh (/ (plist-get (cdr parea) :height) (frame-char-height frame))))
    (erase-buffer)
    (setq mode-line-format nil)
    (insert-plotting-area parea foreground background mouse-1-action)
    (insert (or text "\n"))
    (modify-frame-parameters frame
     `((cursor-type . nil)
       (title . ,(or title "plot"))
       (unsplittable . t)
       (left-fringe . 0)
       (right-fringe . 0)
       (minibuffer . nil)
       (tool-bar-lines . nil)
       (menu-bar-lines . 0)
       (width . ,(1+ fw))
       (height . ,(1+ fh))))
    (current-buffer)))

;; ========== examples:

; (graphs-example-1)
; (graphs-example-2)

(defun graphs-example-1 ()
  (let ((parea (make-plotting-area 800 400 0 5 -1 1)))
    (dotimes (n 15)
      (plot-function (lambda (x) (- (abs (* (cos (sqrt x)) (+ n (sin x)))) (sqrt (/ n 20.0)))) parea))
    (insert-plotting-area parea "blue" "honeydew")))

(defun graphs-example-2 ()
  (let ((parea (make-plotting-area 800 400 0 5 -1 1)))
    (dotimes (n 15)
      (plot-function (lambda (x) (- (abs (* (cos (sqrt x)) (+ n (sin x)))) (sqrt (/ n 10.0)))) parea))
    (pop-up-area parea "blue" "honeydew" nil "un zoli dessin")))

;;======================================================================
;; this is it

(provide 'graphs)

;; graphs.el ends here